// Copyright (c) 2014, The Shelf Route project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_route;

import 'package:shelf/shelf.dart';
import 'package:path/path.dart' as p;

/// A [Handler] that routes to other handlers based on the incoming request
/// (path and method). Note there can be several layers of routers to facilitate
/// modularity.
///
/// When the Router routes a request to a handler it adjusts the requests as follows:
/// * the routes path is removed from the start of the requests pathInfo
/// * the routes path is added to the end of the requests scriptName
///
/// e.g if original request had path info of `/banking' and scriptName was
/// `/abc` then when routing a request of `/banking/accounts` the new request
/// past to the handler will have a pathInfo of `/accounts` and scriptName of
/// `/abc/banking`
///
///
Router router() => new Router();



class Router {
  final List<_Route> _routes = <_Route>[];
  final _Route _fallbackRoute;

  Router._internal(this._fallbackRoute);

  Router([Handler fallbackHandler])
      : this._fallbackRoute = new _Route(fallbackHandler != null ?
          fallbackHandler : _send404, null, null);

  void addRoute(Handler handler, {String path, String method}) {
    _routes.add(new _Route(handler, path, method));
  }

  Router addChildRouter(String path) {
    final child = new Router._internal(_fallbackRoute);
    addRoute(child.handler, path: path);
    return child;
  }

  Handler get handler => _handleRequest;

  _handleRequest(Request request) {
    final route = _routes.firstWhere((r) => r.canHandle(request),
        orElse: () => _fallbackRoute);
    return route.handle(request) ;
  }
}


class _Route {
  final Handler handler;
  final String path;
  final String method;

  _Route(this.handler, String path, this.method)
      : this.path = _stringLeadingSlash(path);

  static String _stringLeadingSlash(String path) {
    if (path == null) {
      return null;
    }

    if (path.isEmpty) {
      throw new ArgumentError('path may not be empty. Use null instead');
    }

    final np = p.normalize(path);
    if (!np.startsWith('/')) {
      return np;
    }

    final result = p.normalize(np.substring(1));
    return result == '.' ? '' : result;
  }

  bool canHandle(Request request) {
    final pathInfo = _stringLeadingSlash(request.pathInfo);

    return
      (path == null || path == pathInfo || p.isWithin(path, pathInfo))
      &&
      (method == null || request.method == method);
  }

  handle(Request request) {
    // create a new request with the pathInfo and scriptName adjusted relative
    // to route path
    final newRequest = _createNewRequest(request);

    return handler(newRequest);
  }

  Request _createNewRequest(Request request) {
    final pathInfo = _stringLeadingSlash(request.pathInfo);
    final scriptName = request.scriptName;

    var newPathInfo = _adjustPath(path == null ? pathInfo :
      p.relative(pathInfo, from: path));

    final newScriptName = _adjustPath(path == null ? scriptName :
      p.join(scriptName, path));

    if (newPathInfo == '' && newScriptName == '') {
      newPathInfo = '/';
    }

//    final newRequest = request.copyWith(pathInfo: newPathInfo,
//        scriptName: newScriptName);

    return new Request(newPathInfo, request.queryString,
        request.method,
        newScriptName, request.protocolVersion, request.requestedUri,
        request.headers);

  }

  String _adjustPath(String path) {
    if (path == '.') {
      path = '';
    }

    if (path.isNotEmpty && !path.startsWith('/')) {
      path = p.join('/', path);
    }

    return path;
  }

}

Response _send404(Request req) {
  return new Response.notFound("Not Found");
}